from snowflake.snowpark.types import StructType, StructField

if not hasattr(StructType, "___extended"):
  StructType.___extended = True    
  def schema__get_item__(self,index):
      if isinstance(index, str):
         name = index.upper()
         for f in self.fields:
            if name == f.name.upper():
               return f
      else:
         return self.fields[index]
  StructType.__getitem__ = schema__get_item__
  StructField.dtype = property(lambda self: self.datatype)


from snowflake.snowpark.types import (
    ArrayType,
    BinaryType,
    BooleanType,
    DataType,
    DateType,
    DecimalType,
    DoubleType,
    GeographyType,
    LongType,
    MapType,
    StringType,
    StructType,
    StructField,
    TimestampType,
    TimeType,
    VariantType,
    FloatType
)
import datetime
import decimal
# based on https://docs.snowflake.com/en/developer-guide/udf/python/udf-python-designing.html#label-sql-python-data-type-mappings

def map_to_python_type_str(dtype:DataType):
    if isinstance(dtype, ArrayType):
        return "list"
    elif isinstance(dtype, BinaryType):
        return "bytes"
    elif isinstance(dtype, BooleanType):
        return "bool"
    elif isinstance(dtype, DateType):
        return "datetime.date"
    elif isinstance(dtype, DecimalType):
        return "int" if dtype.scale == 0 else "decimal.Decimal"
    elif isinstance(dtype, FloatType):
        return "float"
    elif isinstance(dtype, DoubleType):
        return "float"
    elif isinstance(dtype, GeographyType):
        return "dict"
    elif isinstance(dtype, LongType):
        return "int"
    elif isinstance(dtype, MapType):
        return "dict"
    elif isinstance(dtype, StringType):
        return "str"
    elif isinstance(dtype, StructType):
        return "dict"
    elif isinstance(dtype, TimestampType):
        return "datetime.datetime"
    elif isinstance(dtype, TimeType):
        return "datetime.time"
    elif isinstance(dtype, VariantType):
        return "dict"

def map_to_python_type(dtype:DataType):
    if isinstance(dtype, ArrayType):
        return list
    elif isinstance(dtype, BinaryType):
        return bytes
    elif isinstance(dtype, BooleanType):
        return bool
    elif isinstance(dtype, DateType):
        return datetime.date
    elif isinstance(dtype, DecimalType):
        return int if dtype.scale == 0 else decimal.Decimal
    elif isinstance(dtype, FloatType):
        return float
    elif isinstance(dtype, DoubleType):
        return float
    elif isinstance(dtype, GeographyType):
        return dict
    elif isinstance(dtype, LongType):
        return int
    elif isinstance(dtype, MapType):
        return dict
    elif isinstance(dtype, StringType):
        return str
    elif isinstance(dtype, StructType):
        return dict
    elif isinstance(dtype, TimestampType):
        return datetime.datetime
    elif isinstance(dtype, TimeType):
        return datetime.time
    elif isinstance(dtype, VariantType):
        return dict


def map_python_type_to_datatype(type):
    if isinstance(type, list):
        return ArrayType
    elif isinstance(type, bytes):
        return BinaryType
    elif isinstance(type, bool):
        return BooleanType
    elif isinstance(type, datetime.date):
        return DateType
    elif isinstance(type, int):
        return LongType
    elif isinstance(type, float):
        return FloatType
    elif isinstance(type, decimal.Decimal):
        return DecimalType
    elif isinstance(type, dict):
        return MapType
    elif isinstance(type, str):
        return StringType
    elif isinstance(type, datetime.datetime):
        return TimestampType
    elif isinstance(type, datetime.time):
        return TimeType
    else:
        return VariantType

def map_string_type_to_datatype(type):
    type = type.lower()
    if type == "list":
        return ArrayType()
    elif type=="bytes":
        return BinaryType()
    elif type == "bool" or type == "boolean":
        return BooleanType()
    elif type == "date":
        return DateType()
    elif type == "int" or type == "long":
        return LongType()
    elif type == "float":
        return FloatType()
    elif type == "double":
        return DoubleType()
    elif type == "decimal":
        return DecimalType()
    elif type == "dict" or type == "struct":
        return MapType()
    elif type == "str" or type == "string" or type == "text":
        return StringType()
    elif type == "timestamp":
        return TimestampType()
    elif type == "time":
        return TimeType()
    else:
        return VariantType()

def schema_str_to_schema(schema_as_str):
    columns = schema_as_str.split(",")
    schema_fields = []
    for c in columns:
        name, type = c.strip().split(" ")
        datatype = map_string_type_to_datatype(type)
        schema_fields.append(StructField(name,datatype))
    return StructType(schema_fields)

from snowflake.snowpark import DataFrame
def display(data:DataFrame, limit:int=10):
    from IPython.display import display as _display, HTML
    if isinstance(data, DataFrame):
        try:
            output_cell_output = displayHTML
        except:
            output_cell_output = lambda x: _display(HTML(x))
        try:
            print(f"Output restricted to {limit} rows")
            html = data.limit(limit).to_pandas().to_html()
            output_cell_output(html)
        except Exception as e:
            print(f"Error displaying dataframe {e}")
    else:
        _display(data)


from snowflake.snowpark import DataFrame, Row, DataFrameNaFunctions, Column
from snowflake.snowpark.functions import col, lit, udtf, regexp_replace
from snowflake.snowpark import functions as F
from snowflake.snowpark.dataframe import _generate_prefix
from snowflake.snowpark.functions import table_function
from snowflake.snowpark.column import _to_col_if_str, _to_col_if_lit
import pandas as pd
import numpy as np
from snowpark_extensions.utils import map_to_python_type, schema_str_to_schema
import shortuuid
from snowflake.snowpark import context
from snowflake.snowpark.types import StructType,StructField
from snowflake.snowpark._internal.analyzer.expression import Expression, FunctionExpression
from snowflake.snowpark._internal.analyzer.unary_expression import Alias
from snowflake.snowpark._internal.analyzer.analyzer_utils import quote_name

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
    overload,
)
from snowflake.snowpark._internal.type_utils import (
    ColumnOrName,
    ColumnOrSqlExpr,
    LiteralType,
)

if not hasattr(DataFrame,"___extended"):
    DataFrame.___extended = True
    def get_dtypes(schema):
        data = np.array([map_to_python_type(x.datatype) for x in schema.fields])
        # providing an index
        dtypes_series = pd.Series(data, index=schema.names)
        return dtypes_series

    DataFrame.dtypes = property(lambda self: get_dtypes(self.schema))


    def map(self,func,output_types,input_types=None,input_cols=None,to_row=False):
        clazz="map"+shortuuid.uuid()[:8]
        output_schema=[]
        if not input_types:
            input_types = [x.datatype for x in self.schema.fields]
        input_cols_len=len(input_types)
        if not input_cols:
            input_col_names=self.columns[:input_cols_len] 
            _input_cols = [self[x] for x in input_col_names]
        else:
            input_col_names=input_cols
            _input_cols = input_cols
        output_schema = [StructField(f"c_{i+1}",output_types[i]) for i in range(len(output_types))]
        output_cols = [col(x.name) for x in output_schema]
        if to_row:
            myRow = Row(*input_col_names)
            def process_func(self,*argv):
                yield func(myRow(*argv))
        else:
            def process_func(self,*argv):
                yield func(*argv)
        udtf_class = type(clazz, (object, ), {"process":process_func})
        udtf(udtf_class,output_schema=StructType(output_schema), input_types=input_types,name=clazz,replace=True,packages=["snowflake-snowpark-python"])
        return self.join_table_function(clazz,*_input_cols).select(*output_cols)

    DataFrame.map = map

    def simple_map(self,func):
        return self.select(*func(self))

    DataFrame.simple_map = simple_map

    DataFrameNaFunctions.__oldreplace = DataFrameNaFunctions.replace 
    
    def extended_replace(
        self,
        to_replace: Union[
            LiteralType,
            Iterable[LiteralType],
            Dict[LiteralType, LiteralType],
        ],
        value: Optional[Iterable[LiteralType]] = None,
        subset: Optional[Iterable[str]] = None,
        regex=False
    ) -> "snowflake.snowpark.dataframe.DataFrame":
        if regex:
            return self._df.select([regexp_replace(col(x.name), to_replace,value).alias(x.name) if isinstance(x.datatype,StringType) else col(x.name) for x in self._df.schema])
        else:
            return self.__oldreplace(to_replace,value,subset)
    
    DataFrameNaFunctions.replace = extended_replace

    def has_null(col):
        return F.array_contains(F.sql_expr("parse_json('null')"),col) | F.coalesce(F.array_contains(lit(None) ,col),lit(False))

    # SPECIAL COLUMN HELPERS

    
    class SpecialColumn(Column):
        def __init__(self,special_column_base_name="special_column"):
            self.special_col_name = _generate_prefix(special_column_base_name)
            super().__init__(self.special_col_name)
            self._is_special_column = True
            self._expression._is_special_column = True
            sc = self 
            def _expand(df):
                nonlocal sc
                return sc.expand(df)
            self._expression.expand = _expand
            self._special_column_dependencies=[]
            import uuid
            self._hash = hash(str(uuid.uuid4()))
            self.expanded = False
        def __hash__(self):
            return self._hash
        def __eq__(self, other):
            return self._hash == other._hash
        def gen_unique_value_name(self,idx,base_name):
            return base_name if idx == 0 else f"{base_name}_{idx}"
        def add_columns(self,new_cols, alias:str=None):
           pass 
        def expand(self,df):
            self.expanded = True
        @classmethod
        def extract_specials(cls,c):
            if isinstance(c, Expression):
                if c.children:
                    for child in c.children:
                        is_child_special = cls.has_special_column(child)
                        if is_child_special:
                           return True
            if hasattr(c,"_is_special_column"):
                return True
            if hasattr(c,"_expression") and c._expression and c._expression.children:
                  for child in c._expression.children:
                        is_child_special = cls.has_special_column(child)
                        if is_child_special:
                           return True
        @classmethod
        def any_specials(cls,*cols):
            for c in cols:
                for special in cls.specials(c):
                    return True
        @classmethod
        def specials(cls,c):
            """  Returns special columns that might add a join clause to the dataframe """
            if isinstance(c, Expression):
                if c.children:
                    for child in c.children:
                        for special in cls.specials(child):
                            yield special
            elif hasattr(c,"_expression") and c._expression.children:
                for child in c._expression.children:
                        for special in cls.specials(child):
                            yield special
            if hasattr(c,"_special_column_dependencies") and c._special_column_dependencies:
                     for dependency in c._special_column_dependencies:
                        for special in cls.specials(dependency):
                            yield special
            if hasattr(c,"_is_special_column"):
                yield c

    class MapValues(SpecialColumn):
        def __init__(self,array_col):
            super().__init__("values")
            self.array_col = array_col
            self._special_column_dependencies = [array_col]
        def add_columns(self, new_cols, alias:str = None):
            # add itself as column
            new_cols.append(self.alias(alias) if alias else self)
        def expand(self,df):
            if not self.expanded:
                array_col = _to_col_if_str(self.array_col, "values")
                df = df.with_column("__IDX",F.seq8())
                flatten = table_function("flatten")
                seq=_generate_prefix("SEQ")
                key=_generate_prefix("KEY")
                path=_generate_prefix("PATH")
                index=_generate_prefix("INDEX")
                value=_generate_prefix("VALUE")
                this=_generate_prefix("THIS")
                df_values=df.join_table_function(flatten(input=array_col,outer=lit(True)).alias(seq,key,path,index,value,this)).group_by("__IDX").agg(F.array_agg(value).alias(self.special_col_name)).distinct()
                df = df.join(df_values,on="__IDX").drop("__IDX")
                self.expanded = True
            return df
     
    class ArrayFlatten(SpecialColumn):
        def __init__(self,flatten_col,remove_arrays_when_there_is_a_null):
            super().__init__("flatten")
            self.flatten_col = flatten_col
            self.remove_arrays_when_there_is_a_null = remove_arrays_when_there_is_a_null
            self._special_column_dependencies = [flatten_col]
        def add_columns(self, new_cols, alias:str = None):
            new_cols.append(self.alias(alias) if alias else self)
        def expand(self,df):
            if not self.expanded:
                array_col = _to_col_if_str(self.flatten_col, "flatten")
                flatten = table_function("flatten")
                df=df.join_table_function(flatten(array_col).alias("__SEQ_FLATTEN","KEY","PATH","__INDEX_FLATTEN","__FLATTEN_VALUE","THIS"))
                df = df.drop("KEY","PATH","THIS")
                if self.remove_arrays_when_there_is_a_null:
                    df_with_has_null=df.withColumn("__HAS_NULL",has_null(array_col))
                    df_flattened= df_with_has_null.group_by(col("__SEQ_FLATTEN")).agg(F.call_builtin("BOOLOR_AGG",col("__HAS_NULL")).alias("__HAS_NULL"),F.call_builtin("ARRAY_UNION_AGG",col("__FLATTEN_VALUE")).alias("__FLATTEN_VALUE"))
                    df_flattened=df_flattened.with_column("__FLATTEN_VALUE",F.iff("__HAS_NULL", lit(None), col("__FLATTEN_VALUE"))).drop("__HAS_NULL")
                    df=df.drop("__FLATTEN_VALUE").where(col("__INDEX_FLATTEN")==0).join(df_flattened,on="__SEQ_FLATTEN").drop("__SEQ_FLATTEN","__INDEX_FLATTEN").rename("__FLATTEN_VALUE",self.special_col_name)
                else:
                    df_flattened= df.group_by(col("__SEQ_FLATTEN")).agg(F.call_builtin("ARRAY_UNION_AGG",col("__FLATTEN_VALUE")).alias("__FLATTEN_VALUE"))
                    df=df.drop("__FLATTEN_VALUE").where(col("__INDEX_FLATTEN")==0).join(df_flattened,on="__SEQ_FLATTEN").drop("__SEQ_FLATTEN","__INDEX_FLATTEN").rename("__FLATTEN_VALUE",self.special_col_name)
                self.expanded = True
            return df

    class ArrayZip(SpecialColumn):
        def __init__(self,left,*right,use_compat=False):
            super().__init__("zipped")
            self.left_col  = left
            self.right_cols = right
            self._special_column_dependencies = [left,*right]
            self._use_compat = use_compat
        def add_columns(self,new_cols,alias:str = None):
            new_cols.append(self.alias(alias) if alias else self)
        def expand(self,df):
            if not self.expanded:
                if (not hasattr(ArrayZip,"_added_compat_func")) and self._use_compat:
                    context.get_active_session().sql("""
    CREATE OR REPLACE TEMPORARY FUNCTION PUBLIC.ARRAY_UNDEFINED_COMPACT(ARR VARIANT) RETURNS ARRAY
    LANGUAGE JAVASCRIPT AS
    $$
        if (ARR.includes(undefined)){
            filtered = ARR.filter(x => x === undefined);
            if (filtered.length==0)
                return filtered;
        }
        return ARR;
    $$;                
                    """).count()
                    ArrayZip._added_compat_func = True
                df_with_idx = df.with_column("_IDX",F.seq8())
                flatten = table_function("flatten")
                right = df_with_idx.select("_IDX",self.left_col)\
                    .join_table_function(flatten(input=self.left_col,outer=lit(True))\
                    .alias("SEQ","KEY","PATH","INDEX","__VALUE_0","THIS")) \
                    .drop(self.left_col,"SEQ","KEY","PATH","THIS")
                vals=["__VALUE_0"]
                for right_col in self.right_cols:
                    prior=len(vals)-1
                    next=len(vals)
                    left_col_name=f"__VALUE_{prior}"
                    right_col_name=f"__VALUE_{next}"
                    vals.append(right_col_name)
                    new_right=df_with_idx.select("_IDX",right_col).join_table_function(flatten(input=right_col,outer=lit(False))\
                        .alias("SEQ","KEY","PATH","INDEX",right_col_name,"THIS")) \
                        .drop(right_col,"SEQ","KEY","PATH","THIS") #.with_column("INDEX",F.coalesce(col("INDEX"),lit(0))) \
                    if right:
                        right = right.join(new_right,on=["_IDX","INDEX"],how="left",lsuffix="___LEFT")
                    else:
                        right = new_right
                zipped = right.select("_IDX","INDEX",F.array_construct(*vals).alias("NGROUP"))
                if self._use_compat:
                    zipped = zipped.with_column("NGROUP",F.call_builtin("ARRAY_UNDEFINED_COMPACT",col("NGROUP")))  
                zipped=zipped.group_by("_IDX").agg(F.sql_expr(f'arrayagg(ngroup) within group (order by INDEX) {self.special_col_name}'))
                result = df_with_idx.join(zipped,on="_IDX").drop("_IDX")
                df = result
                self.expanded = True
            return df
        
    class Explode(SpecialColumn):
        def __init__(self,expr,map=False,outer=False,use_compat=False):
            """ Right not it must be explictly stated if the value is a map. By default it is assumed it is not"""
            super().__init__("value" if map else "col")
            self.expr = expr
            self.map = map
            self.outer = outer
            self.key_col_name = None
            self.key_col = None
            self.value_col_name = self.special_col_name
            self._special_column_dependencies = [expr]
            self.use_compat=use_compat
        def add_columns(self,new_cols,alias:str = None):
            if self.map:
                self.key_col_name = _generate_prefix("key")
                self.key_col = col(self.key_col_name)
                new_cols.append(self.key_col.alias(alias + "_1") if alias else self.key_col)
            new_cols.append(self.alias(alias) if alias else self)
        def expand(self,df):
            if not self.expanded:
                if self.key_col_name is not None:
                    df = df.join_table_function(flatten(input=self.expr,outer=lit(self.outer)).alias("SEQ",self.key_col_name,"PATH","INDEX",self.value_col_name,"THIS")).drop(["SEQ","PATH","INDEX","THIS"])
                else:
                    df = df.join_table_function(flatten(input=self.expr,outer=lit(self.outer)).alias("SEQ","KEY","PATH","INDEX",self.value_col_name,"THIS")).drop(["SEQ","KEY","PATH","INDEX","THIS"])
                    if self.use_compat:
                        df=df.with_column(self.value_col_name,
                        F.iff(
                            F.cast(self.value_col_name,ArrayType()) == F.array_construct(),
                            lit(None),
                            F.cast(self.value_col_name,ArrayType())))
                self.expanded = True
            return df

    def explode(expr,outer=False,map=False,use_compat=False):
        return Explode(expr,map,outer,use_compat=use_compat)

    def explode_outer(expr,map=False, use_compat=False):
        return Explode(expr,map,True,use_compat=use_compat)

    F.explode = explode
    F.explode_outer = explode_outer
    def _arrays_zip(left,*right,use_compat=False):
        """ In SF zip might return [undefined,...undefined] instead of [] """
        return ArrayZip(left,*right,use_compat=use_compat)
    def _arrays_flatten(array_col,remove_arrays_when_there_is_a_null=True):
        return ArrayFlatten(array_col,remove_arrays_when_there_is_a_null)
    def _map_values(col:ColumnOrName):
        col = _to_col_if_str(col,"map_values")
        return MapValues(col)

    F.map_values = _map_values    
    F.arrays_zip = _arrays_zip
    F.flatten    = _arrays_flatten
    flatten = table_function("flatten")
    _oldwithColumn = DataFrame.withColumn
    _oldSelect = DataFrame.select
    def withColumnExtended(self,colname,expr):
        if isinstance(expr, SpecialColumn):
            new_cols = []
            expr.add_columns(new_cols, alias=colname)
            df = self
            for s in SpecialColumn.specials(expr):
                df=s.expand(df)
            return _oldSelect(df,*self.columns,*new_cols)
            #return self.with_columns(df,new_cols,[col(x) for x in new_cols])
        else:
            return _oldwithColumn(self,colname,expr)
        
    DataFrame.withColumn = withColumnExtended


    def selectExtended(self,*cols):
        if SpecialColumn.any_specials(*cols):
          new_cols = []
          extended_cols = []
          # extend only the main cols
          for x in cols:
            if isinstance(x, SpecialColumn):
                x.add_columns(new_cols)
                extended_cols.append(x)
            else:
                new_cols.append(x)
          df = self
          # but expand all tables, because there could be several joins
          for c in cols:
            for extended_col in SpecialColumn.specials(c):
                df = extended_col.expand(df)
          return _oldSelect(df,*[_to_col_if_str(x,"extended") for x in new_cols])
        else:
            return _oldSelect(self,*cols)
    
    DataFrame.select = selectExtended


import shortuuid
from snowflake.snowpark import Window, Column
from snowflake.snowpark.types import *
from snowflake.snowpark.functions import udtf, col
from snowflake.snowpark.relational_grouped_dataframe import RelationalGroupedDataFrame

def group_by_pivot(self,pivot_col):
      return GroupByPivot(self, pivot_col)
RelationalGroupedDataFrame.pivot = group_by_pivot

class GroupByPivot():
      def __init__(self,old_groupby_col,pivot_col):
            self.old_groupby_col = old_groupby_col
            self.pivot_col=pivot_col
            group_by_exprs = [F.sql_expr(x.sql) for x in old_groupby_col._grouping_exprs]
            group_by_exprs.append(pivot_col)
            self.df = old_groupby_col._df.groupBy(group_by_exprs)
      def clean(self,pivoted):
        def get_valid_id(id):
            return id if re.match("[A-Za-z]\w+", id) else f'"{id}"'
        n = len(self.value_list)
        last_n_cols = self.value_list[-n:]
        pivoted_cols = pivoted.columns[-n:]
        previous_cols = pivoted.columns[:len(pivoted.columns)-n]
        for i in range(0,n):
            the_col = pivoted_cols[i]
            the_value = self.value_list[i]
            pivoted_cols[i] = col(the_col).alias(get_valid_id(str(the_value)))
        return pivoted.select(*previous_cols,*pivoted_cols)
      def prepare(self,col):
            first = self.df.agg(col.alias("__firstAggregate"))
            self.value_list = [x[0] for x in first.select(self.pivot_col).distinct().sort(self.pivot_col).collect()]
            return first.pivot(pivot_col=self.pivot_col,values=self.value_list)
      def avg(self, col: ColumnOrName) -> DataFrame:
            """Return the average for the specified numeric columns."""
            return self.prepare(F.avg(col)).avg("__firstAggregate")
      mean = avg
      def sum(self, col: ColumnOrName) -> DataFrame:
            """Return the sum for the specified numeric columns."""
            return  self.clean(self.prepare(F.sum(col)).sum("__firstAggregate"))
      def median(self, col: ColumnOrName) -> DataFrame:
            """Return the median for the specified numeric columns."""
            return self.clean(self.prepare(F.median(col)).median("__firstAggregate"))
      def min(self, col: ColumnOrName) -> DataFrame:
            """Return the min for the specified numeric columns."""
            return self.clean(self.prepare(F.min(col)).min("__firstAggregate"))
      def max(self, col: ColumnOrName) -> DataFrame:
            """Return the max for the specified numeric columns."""
            return self.clean(self.prepare(col).max("__firstAggregate"))
      def count(self) -> DataFrame:
            """Return the number of rows for each group."""
            return self.clean(self.prepare(col).count("__firstAggregate"))
      def agg(self, aggregated_col: ColumnOrName) -> DataFrame:       
            if hasattr(aggregated_col, "_expression") and isinstance(aggregated_col._expression, FunctionExpression):
                name = aggregated_col._expression.name
                return self.clean(self.prepare(aggregated_col).function(name)(col("__firstAggregate")))
            else:
                raise Exception("Also functions expressions are supported") 

if not hasattr(RelationalGroupedDataFrame, "applyInPandas"):
  def applyInPandas(self,func,schema):
      output_schema = schema
      if isinstance(output_schema, str):
        output_schema = schema_str_to_schema(output_schema)
      from snowflake.snowpark.functions import col
      input_types = [x.datatype for x in self._df.schema.fields]
      input_cols  = [x.name for x in self._df.schema.fields]
      output_cols = [x.name for x in output_schema.fields]
      grouping_exprs = [Column(x) for x in self._grouping_exprs]
      clazz="applyInPandas"+shortuuid.uuid()[:8]
      def __init__(self):
          self.rows = []
          self.dfs  = []
      def process(self, *row):
          self.rows.append(row)
          # Merge rows into a dataframe
          if len(self.rows) >= 16000:
             df = pd.DataFrame(self.rows)
             self.dfs.append(df)
             self.rows = []
          # Merge dataframes into a single dataframe
          if len(self.dfs) >= 100:
             merged_df = pd.concat(self.dfs)
             self.dfs = [merged_df]
          yield None
      def end_partition(self):
        # Merge any remaining rows
        if len(self.rows) > 0:
          df = pd.DataFrame(self.rows,columns=input_cols)
          self.dfs.append(df)
          self.rows = []
        pandas_input = pd.concat(self.dfs)        
        pandas_output = func(pandas_input)
        for row in pandas_output.itertuples(index=False):
             yield tuple(row)
      non_ambigous_output_schema = StructType([StructField(f"pd_{i}",output_schema.fields[i].datatype) for i in range(len(output_schema.fields))])
      renamed_back = [col(f"pd_{i}").alias(output_schema.fields[i].name) for i in range(len(non_ambigous_output_schema.fields))]
      udtf_class = type(clazz, (object, ), {"__init__":__init__,"process":process,"end_partition":end_partition})
      tfunc = udtf(udtf_class,output_schema=non_ambigous_output_schema, input_types=input_types,name=clazz,replace=True,is_permanent=False,packages=["snowflake-snowpark-python", "pandas"])
      return self._df.join_table_function(tfunc(*input_cols).over(partition_by=grouping_exprs, order_by=grouping_exprs)).select(*renamed_back)

  RelationalGroupedDataFrame.applyInPandas = applyInPandas
  ###### HELPER END   






from snowflake.snowpark import functions as F
from snowflake.snowpark import context
from snowflake.snowpark.functions import call_builtin, col,lit, concat, coalesce, object_construct_keep_null
from snowflake.snowpark import DataFrame, Column
from snowflake.snowpark.types import ArrayType, BooleanType
from snowflake.snowpark._internal.type_utils import (
    ColumnOrLiteral,
    ColumnOrLiteralStr,
    ColumnOrName,
    ColumnOrSqlExpr,
    LiteralType,
)
from snowflake.snowpark.column import _to_col_if_str, _to_col_if_lit
from snowflake.snowpark.dataframe import _generate_prefix
from snowflake.snowpark._internal.analyzer.unary_expression import Alias

if not hasattr(F,"___extended"):
    F.___extended = True

    def pairwise(iterable):
        while len(iterable):
            a = iterable.pop(0)
            if len(iterable):
                b = iterable.pop(0)
            else:
                b = None
            yield (a,b)
    
    def flatten_col_list(obj):
        if isinstance(obj, str) or isinstance(obj, Column):
            return [obj]
        elif hasattr(obj, '__iter__'):
            acc = []
            for innerObj in obj:
                acc = acc + flatten_col_list(innerObj)
            return acc


    def regexp_extract(value:ColumnOrLiteralStr,regexp:ColumnOrLiteralStr,idx:int) -> Column:
        """
        Extract a specific group matched by a regex, from the specified string column. 
        If the regex did not match, or the specified group did not match, 
        an empty string is returned.        
        """
        value = _to_col_if_str(value,"regexp_extract")
        regexp = _to_col_if_lit(regexp,"regexp_extract")
        idx = _to_col_if_lit(idx,"regexp_extract")
        # we add .* to the expression if needed
        return coalesce(call_builtin('regexp_substr',value,regexp,lit(1),lit(1),lit('e'),idx),lit(''))

    def unix_timestamp(col):
        return call_builtin("DATE_PART","epoch_second",col)

    def from_unixtime(col):
        col = _to_col_if_str(col,"from_unixtime")
        return F.to_timestamp(col).alias('ts')

    def format_number(col,d):
        col = _to_col_if_str(col,"format_number")
        return F.to_varchar(col,'999,999,999,999,999.' + '0'*d)

    def reverse(col):
        col = _to_col_if_str(col,"reverse")
        return F.call_builtin('reverse',col)

    def date_add(col,num_of_days):
        col = _to_col_if_str(col,"date_add")
        num_of_days=_to_col_if_str_or_int(num_of_days)
        return dateadd(lit('day'),col,num_of_days)

    def date_sub(col,num_of_days):
        col = _to_col_if_str(col,"date_sub")
        num_of_days=_to_col_if_str_or_int(num_of_days)
        return dateadd(lit('day'),col,-1 * num_of_days)

    def create_map(*col_names):
        """
        Usage:
        res = df.select(create_map('name', 'age').alias("map")).collect()
        """
        from snowflake.snowpark.functions import col,lit, object_construct
        col_list = []
        # flatten any iterables, to process them in pairs
        col_names = flatten_col_list(col_names)
        for name, value in pairwise(col_names):
            if isinstance(name, str):
                col_list.append(lit(name))
            else:
                col_list.append(name)
            col_list.append(value)
        return object_construct(*col_list)

    def array_distinct(col):
        col = _to_col_if_str(col,"array_distinct")
        return F.call_builtin('array_distinct',col)


    def _array(*cols):
        return F.array_construct(*cols)

    F._sort_array_function = None
    def _sort_array(col:ColumnOrName,asc:ColumnOrLiteral=True):
        if not F._sort_array_function:
            session = context.get_active_session()
            current_database = session.get_current_database()
            function_name =_generate_prefix("_sort_array_helper")
            F._sort_array_function = f"{current_database}.public.{function_name}"
            session.sql(f"""
            create or replace temporary function {F._sort_array_function}(ARR ARRAY,ASC BOOLEAN) returns ARRAY
            language javascript as
            $$
            ARRLENGTH = ARR.length;
            // filter nulls
            ARR = ARR.filter(x => x !== null);
            if (ARR.length && ARR[0] instanceof Object)
            {{
                function sortFn() 
                {{
                    var sortByProps = Array.prototype.slice.call(arguments),
                        cmpFn = function(left, right, sortOrder) {{
                            var sortMultiplier = sortOrder === "asc" ? 1 : -1;
                            if (left > right) {{ return +1 * sortMultiplier;}}
                            if (left < right) {{ return -1 * sortMultiplier;}}
                            return 0;
                        }};
                    return function(sortLeft, sortRight) {{
                        // get value from object by complex key
                        var getValueByStr = function(obj, path) {{
                        var i, len;
                        //prepare keys
                        path = path.replace('[', '.');
                        path = path.replace(']', '');
                        path = path.split('.');
                        len = path.length;
                        for (i = 0; i < len; i++) {{
                        if (!obj || typeof obj !== 'object') {{ return obj;}}
                        obj = obj[path[i]];
                        }}
                return obj;
                }};
                return sortByProps.map(function(property) {{
                    return cmpFn(getValueByStr(sortLeft, property.prop), getValueByStr(sortRight, property.prop), property.sortOrder);
                }}).reduceRight(function(left, right) {{
                    return right || left;
            }});
        }};
        }}
        var props = Object.getOwnPropertyNames(ARR[0]);
        var sortKeys = [];
        for(var p of props)
        {{
            sortKeys.push({{prop:p,sortOrder:"asc"}});
        }}
        ARR.sort(sortFn(...sortKeys));
        }}
        else
            ARR.sort();
            var RES = new Array(ARRLENGTH-ARR.length).fill(null).concat(ARR);
            if (ASC) return RES; else return RES.reverse();
        $$;""").show()
        return call_builtin(F._sort_array_function,col,asc)


    F._array_sort_function = None
    def _array_sort(col:ColumnOrName):
        if not F._array_sort_function:
            session = context.get_active_session()
            current_database = session.get_current_database()
            function_name =_generate_prefix("_array_sort_helper")
            F._array_sort_function = f"{current_database}.public.{function_name}"
            session.sql(f"""
            create or replace temporary function {F._array_sort_function}(ARR ARRAY) returns ARRAY
            language javascript as
            $$
            ARRLENGTH = ARR.length;
            // filter nulls
            ARR = ARR.filter(x => x !== null);
            if (ARR.length && ARR[0] instanceof Object)
            {{
                function sortFn() 
                {{
                    var sortByProps = Array.prototype.slice.call(arguments),
                        cmpFn = function(left, right, sortOrder) {{
                            var sortMultiplier = sortOrder === "asc" ? 1 : -1;
                            if (left > right) {{ return +1 * sortMultiplier;}}
                            if (left < right) {{ return -1 * sortMultiplier;}}
                            return 0;
                        }};
                    return function(sortLeft, sortRight) {{
                        // get value from object by complex key
                        var getValueByStr = function(obj, path) {{
                        var i, len;
                        //prepare keys
                        path = path.replace('[', '.');
                        path = path.replace(']', '');
                        path = path.split('.');
                        len = path.length;
                        for (i = 0; i < len; i++) {{
                        if (!obj || typeof obj !== 'object') {{ return obj;}}
                        obj = obj[path[i]];
                        }}
                return obj;
                }};
                return sortByProps.map(function(property) {{
                    return cmpFn(getValueByStr(sortLeft, property.prop), getValueByStr(sortRight, property.prop), property.sortOrder);
                }}).reduceRight(function(left, right) {{
                    return right || left;
            }});
        }};
        }}
        var props = Object.getOwnPropertyNames(ARR[0]);
        var sortKeys = [];
        for(var p of props)
        {{
            sortKeys.push({{prop:p,sortOrder:"asc"}});
        }}
        ARR.sort(sortFn(...sortKeys));
        }}
        else
            ARR.sort();
        var RES = ARR.concat(new Array(ARRLENGTH-ARR.length).fill(null));
        return RES;
        $$;""").show()
        return call_builtin(F._array_sort_function,col)        
    F._array_max_function = None
    def _array_max(col:ColumnOrName):
        if not F._sort_array_function:
            session = context.get_active_session()
            current_database = session.get_current_database()
            function_name =_generate_prefix("_array_max_function")
            F._array_max_function = f"{current_database}.public.{function_name}"
            session.sql(f"""
            create or replace temporary function {F._array_max_function}(ARR ARRAY) returns VARIANT
            language javascript as
            $$
            return Math.max(...ARR);
            $$
            """).show()
        return call_builtin(F._array_max_function,col)
    F._array_min_function = None
    def _array_min(col:ColumnOrName):
        if not F._sort_array_function:
            session = context.get_active_session()
            current_database = session.get_current_database()
            function_name =_generate_prefix("_array_min_function")
            F._array_min_function = f"{current_database}.public.{function_name}"
            session.sql(f"""
            create or replace temporary function {F._array_min_function}(ARR ARRAY) returns VARIANT
            language javascript as
            $$
            return Math.min(...ARR);
            $$
            """).show()
        return call_builtin(F._array_min_function,col)

    def _struct(*cols):
        new_cols = []
        for c in flatten_col_list(cols):
            if isinstance(c, str):
                new_cols.append(lit(c))
            else:
                name = c._expression.name
                name = name[1:] if name.startswith('"') else name
                name = name[:-1] if name.endswith('"') else name
                new_cols.append(lit(name))
            c = _to_col_if_str(c, "struct")
            if isinstance(c, Column) and isinstance(c._expression,Alias):
                new_cols.append(col(c._expression.children[0])) 
            else:
                new_cols.append(c)
        return object_construct_keep_null(*new_cols)

    def _bround(col: Column, scale: int = 0): 
        power = pow(F.lit(10), F.lit(scale))
        elevatedColumn = F.when(F.lit(0) == F.lit(scale), col).otherwise(col * power)
        columnFloor = F.floor(elevatedColumn)
        return F.when(
            elevatedColumn - columnFloor == F.lit(0.5)
            , F.when(columnFloor % F.lit(2) == F.lit(0), columnFloor).otherwise(columnFloor + F.lit(1))
        ).otherwise(F.round(elevatedColumn)) / F.when(F.lit(0) == F.lit(scale), F.lit(1)).otherwise(power)
    

    F.array = _array
    F.array_max = _array_max
    F.array_min = _array_min
    F.array_distinct = array_distinct
    F.regexp_extract = regexp_extract
    F.create_map = create_map
    F.unix_timestamp = unix_timestamp
    F.from_unixtime = from_unixtime
    F.format_number = format_number
    F.reverse = reverse
    F.date_add = date_add
    F.date_sub = date_sub
    F.asc  = lambda col: _to_col_if_str(col, "asc").asc()
    F.desc = lambda col: _to_col_if_str(col, "desc").desc()
    F.asc_nulls_first = lambda col: _to_col_if_str(col, "asc_nulls_first").asc()
    F.desc_nulls_first = lambda col: _to_col_if_str(col, "desc_nulls_first").asc()
    F.sort_array = _sort_array
    F.array_sort = _array_sort
    F.struct = _struct
    F.bround = _bround